from orchestrator.db.sorting.sorting import Sort, SortOrder, generic_apply_sorts, generic_sort, generic_sorts_validate

__all__ = [
    "Sort",
    "SortOrder",
    "generic_sort",
    "generic_apply_sorts",
    "generic_sorts_validate",
]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import structlog
from graphql import GraphQLError
from sqlalchemy.orm import defer, joinedload

from orchestrator.db import ProcessSubscriptionTable, ProcessTable, SubscriptionTable
from orchestrator.db.filters import CallableErrorHander, Filter
from orchestrator.db.filters.process import filter_processes
from orchestrator.db.range import apply_range_to_query
from orchestrator.db.sorting import Sort
from orchestrator.db.sorting.process import sort_processes
from orchestrator.graphql.pagination import Connection, PageInfo
from orchestrator.graphql.schemas.process import ProcessGraphqlSchema, ProcessType
from orchestrator.graphql.types import CustomInfo, GraphqlFilter, GraphqlSort
from orchestrator.services.processes import load_process
from orchestrator.utils.show_process import show_process

logger = structlog.get_logger(__name__)


def enrich_process(process: ProcessTable) -> ProcessGraphqlSchema:
    p = load_process(process)
    data = show_process(process, p)
    return ProcessGraphqlSchema(**data)


def handle_process_error(info: CustomInfo) -> CallableErrorHander:
    def _handle_process_error(message: str, **kwargs) -> None:  # type: ignore
        logger.debug(message, **kwargs)
        extra_values = dict(kwargs.items()) if kwargs else {}
        info.context.errors.append(GraphQLError(message=message, path=info.path, extensions=extra_values))

    return _handle_process_error


async def resolve_processes(
    info: CustomInfo,
    filter_by: Union[list[GraphqlFilter], None] = None,
    sort_by: Union[list[GraphqlSort], None] = None,
    first: int = 10,
    after: int = 0,
) -> Connection[ProcessType]:
    _error_handler = handle_process_error(info)

    _range: Union[list[int], None] = [after, after + first] if after is not None and first else None
    pydantic_filter_by: list[Filter] = [item.to_pydantic() for item in filter_by] if filter_by else []
    pydantic_sort_by: list[Sort] = [item.to_pydantic() for item in sort_by] if sort_by else []
    logger.info("resolve_processes() called", range=_range, sort=sort_by, filter=pydantic_filter_by)

    # the joinedload on ProcessSubscriptionTable.subscription via ProcessBaseSchema.process_subscriptions prevents a query for every subscription later.
    # tracebacks are not presented in the list of processes and can be really large.
    query = ProcessTable.query.options(
        joinedload(ProcessTable.process_subscriptions)
        .joinedload(ProcessSubscriptionTable.subscription)
        .joinedload(SubscriptionTable.product),
        defer("traceback"),
    )

    query = filter_processes(query, pydantic_filter_by, _error_handler)
    query = sort_processes(query, pydantic_sort_by, _error_handler)
    total = query.count()
    query = apply_range_to_query(query, after, first)

    processes = query.all()
    has_next_page = len(processes) > first

    # exclude last item as it was fetched to know if there is a next page
    processes = processes[:first]
    processes_length = len(processes)
    start_cursor = after if processes_length else None
    end_cursor = after + processes_length - 1
    page_processes = [ProcessType.from_pydantic(enrich_process(p)) for p in processes]

    return Connection(
        page=page_processes,
        page_info=PageInfo(
            has_previous_page=bool(after),
            has_next_page=has_next_page,
            start_cursor=start_cursor,
            end_cursor=end_cursor,
            total_items=total if total else None,
        ),
    )



from datetime import datetime
from typing import Any, Dict, List, Optional, Union
from uuid import UUID

import strawberry
from strawberry.scalars import JSON

from orchestrator.config.assignee import Assignee
from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.schemas.process import ProcessForm, ProcessStepSchema
from orchestrator.workflow import ProcessStatus


# TODO: Change to the orchestrator.schemas.process version when subscriptions are typed in strawberry.
class ProcessBaseSchema(OrchestratorBaseModel):
    id: UUID
    workflow_name: str
    product: Optional[UUID]
    customer: Optional[UUID]
    assignee: Assignee
    failed_reason: Optional[str]
    traceback: Optional[str]
    step: Optional[str]
    status: ProcessStatus
    last_step: Optional[str]
    created_by: Optional[str]
    started: datetime
    last_modified: datetime
    # subscriptions: List[SubscriptionBaseSchema]
    is_task: bool


class ProcessGraphqlSchema(ProcessBaseSchema):
    current_state: Dict[str, Any]
    steps: List[ProcessStepSchema]
    form: Optional[ProcessForm]


@strawberry.experimental.pydantic.type(model=ProcessForm)
class ProcessFormType:
    title: strawberry.auto
    type: strawberry.auto
    additionalProperties: strawberry.auto
    required: strawberry.auto
    properties: JSON
    definitions: Union[JSON, None]


@strawberry.experimental.pydantic.type(model=ProcessStepSchema)
class ProcessStepType:
    stepid: strawberry.auto
    name: strawberry.auto
    status: strawberry.auto
    created_by: strawberry.auto
    executed: strawberry.auto
    commit_hash: strawberry.auto
    state: Union[JSON, None]


@strawberry.experimental.pydantic.type(model=ProcessGraphqlSchema)
class ProcessType:
    id: strawberry.auto
    workflow_name: strawberry.auto
    product: strawberry.auto
    customer: strawberry.auto
    assignee: strawberry.auto
    failed_reason: strawberry.auto
    traceback: strawberry.auto
    step: strawberry.auto
    status: strawberry.auto
    last_step: strawberry.auto
    created_by: strawberry.auto
    started: strawberry.auto
    last_modified: strawberry.auto
    is_task: strawberry.auto
    steps: strawberry.auto
    form: strawberry.auto
    current_state: Union[JSON, None]



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import Callable
from uuid import UUID

import structlog
from sqlalchemy import String, cast

from orchestrator.api.error_handling import raise_status
from orchestrator.db import ProcessSubscriptionTable, ProcessTable, ProductTable, SubscriptionTable, db
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters.filters import generic_filter

logger = structlog.get_logger(__name__)


def pid_filter(query: SearchQuery, value: str) -> SearchQuery:
    return query.filter(cast(ProcessTable.pid, String).ilike("%" + value + "%"))


def is_task_filter(query: SearchQuery, value: str) -> SearchQuery:
    value_as_bool = value.lower() in ("yes", "y", "ye", "true", "1", "ja")
    return query.filter(ProcessTable.is_task.is_(value_as_bool))


def assignee_filter(query: SearchQuery, value: str) -> SearchQuery:
    assignees = value.split("-")
    return query.filter(ProcessTable.assignee.in_(assignees))


def status_filter(query: SearchQuery, value: str) -> SearchQuery:
    statuses = value.split("-")
    return query.filter(ProcessTable.last_status.in_(statuses))


def workflow_filter(query: SearchQuery, value: str) -> SearchQuery:
    return query.filter(ProcessTable.workflow.ilike("%" + value + "%"))


def creator_filter(query: SearchQuery, value: str) -> SearchQuery:
    return query.filter(ProcessTable.created_by.ilike("%" + value + "%"))


def organisation_filter(query: SearchQuery, value: str) -> SearchQuery:
    try:
        value_as_uuid = UUID(value)
    except (ValueError, AttributeError):
        msg = f"Not a valid organisation, must be a UUID: '{value}'"
        logger.debug(msg)
        raise_status(HTTPStatus.BAD_REQUEST, msg)

    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .filter(SubscriptionTable.customer_id == value_as_uuid)
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def product_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable, ProductTable)
        .filter(ProductTable.name.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def tag_filter(query: SearchQuery, value: str) -> SearchQuery:
    tags = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable, ProductTable)
        .filter(ProductTable.tag.in_(tags))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def subscriptions_filter(query: SearchQuery, value: str) -> SearchQuery:
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .join(SubscriptionTable)
        .filter(SubscriptionTable.description.ilike("%" + value + "%"))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


def target_filter(query: SearchQuery, value: str) -> SearchQuery:
    targets = value.split("-")
    process_subscriptions = (
        db.session.query(ProcessSubscriptionTable)
        .filter(ProcessSubscriptionTable.workflow_target.in_(targets))
        .subquery()
    )
    return query.filter(ProcessTable.pid == process_subscriptions.c.pid)


VALID_FILTER_FUNCTIONS_BY_COLUMN: dict[str, Callable[[SearchQuery, str], SearchQuery]] = {
    "pid": pid_filter,
    "istask": is_task_filter,
    "assignee": assignee_filter,
    "status": status_filter,
    "workflow": workflow_filter,
    "creator": creator_filter,
    "organisation": organisation_filter,
    "product": product_filter,
    "tag": tag_filter,
    "subscription": subscriptions_filter,
    "target": target_filter,
}


filter_processes = generic_filter(VALID_FILTER_FUNCTIONS_BY_COLUMN)



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Generator

from graphql import GraphQLError
from strawberry.extensions import SchemaExtension

from orchestrator.graphql.types import CustomInfo


class ErrorCollectorExtension(SchemaExtension):
    errors: list[GraphQLError] = []

    def resolve(self, _next: Any, root: Any, info: CustomInfo, *args, **kwargs) -> Any:  # type: ignore
        info.context.errors = self.errors
        return _next(root, info, *args, **kwargs)

    def on_execute(self) -> Generator[None, None, None]:
        self.errors.clear()
        yield
        if self.execution_context.result:
            if not self.execution_context.result.errors:
                self.execution_context.result.errors = []
            self.execution_context.result.errors.extend(self.errors)



from orchestrator.db import ProcessTable
from orchestrator.db.sorting.sorting import generic_sort

VALID_SORT_KEY_MAP = {
    "creator": "created_by",
    "started": "started_at",
    "status": "last_status",
    "assignee": "assignee",
    "modified": "last_modified_at",
    "workflow": "workflow",
}

sort_processes = generic_sort(VALID_SORT_KEY_MAP, ProcessTable)



# Copyright 2022-2023 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Union

import strawberry
import structlog
from fastapi import HTTPException
from graphql.pyutils import Path
from strawberry import BasePermission
from strawberry.types import Info
from strawberry.types.fields.resolver import StrawberryResolver

from orchestrator.settings import app_settings, oauth2_settings

logger = structlog.get_logger(__name__)

# Note: if this becomes generic enough we might want to move this to oauth2_lib, so it can be reused also by
# other graphql providers (nwd-api, CIM, etc.)


def get_path_as_string(path: Path) -> str:
    if path.prev:
        return f"{get_path_as_string(path.prev)}/{path.key}"
    else:
        return f"{path.key}"


class IsAuthenticated(BasePermission):
    message = "User is not authenticated"

    async def has_permission(self, source: Any, info: Info, **kwargs) -> bool:  # type: ignore
        if not oauth2_settings.OAUTH2_ACTIVE:
            return True

        path = f"/{app_settings.SERVICE_NAME}/{get_path_as_string(info.path)}/".lower()

        context = info.context
        try:
            logger.debug("Request headers", headers=info.context.request.headers)
            current_user = await context.get_current_user(info.context.request)
        except HTTPException:
            self.message = f"User is not authorized to query or has an invalid access token for path: `{path}`"
            return False

        opa_decision = await context.get_opa_decision(path, current_user)

        logger.debug("Get opa decision", path=path, opa_decision=opa_decision)
        if not opa_decision:
            self.message = f"User is not authorized to query `{path}`"

        return opa_decision


class IsAuthenticatedForMutation(BasePermission):
    message = "User is not authenticated"

    async def has_permission(self, source: Any, info: Info, **kwargs) -> bool:  # type: ignore
        mutations_active = oauth2_settings.OAUTH2_ACTIVE and app_settings.MUTATIONS_ENABLED

        if not mutations_active:
            return app_settings.ENVIRONMENT in app_settings.ENVIRONMENT_IGNORE_MUTATION_DISABLED

        path = f"{app_settings.SERVICE_NAME}/{info.path.key}"
        try:
            current_user = await info.context.get_current_user(info.context.request)
        except HTTPException:
            self.message = f"User is not authorized to query or has an invalid access token for path: `{path}`"
            return False
        opa_decision: bool = await info.context.get_opa_decision(path, current_user)

        logger.debug("Get opa decision", path=path, opa_decision=opa_decision)
        if not opa_decision:
            self.message = f"User is not authorized to execute mutation `{path}`"

        return opa_decision


def authenticated_field(
    description: str,
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod, None] = None,
    deprecation_reason: Union[str, None] = None,
) -> Any:
    return strawberry.field(
        description=description,
        resolver=resolver,  # type: ignore
        deprecation_reason=deprecation_reason,
        permission_classes=[IsAuthenticated],
    )


def authenticated_federated_field(  # type: ignore
    description: str,
    resolver: Union[StrawberryResolver, Callable, staticmethod, classmethod, None] = None,
    deprecation_reason: Union[str, None] = None,
    requires: Union[list[str], None] = None,
    **kwargs,
) -> Any:
    return strawberry.federation.field(
        description=description,
        resolver=resolver,  # type: ignore
        deprecation_reason=deprecation_reason,
        permission_classes=[IsAuthenticated],
        requires=requires,
        **kwargs,
    )



"""(regression)tests in relation to domain model caching."""
from http import HTTPStatus
from os import getenv
from unittest.mock import patch

import pytest
from nwastdlib.url import URL
from redis import Redis

from orchestrator.domain.base import SubscriptionModel
from orchestrator.services.subscriptions import build_extended_domain_model
from orchestrator.settings import app_settings
from orchestrator.utils.redis import to_redis


@pytest.fixture
def cache_fixture():
    # Fixture to enable domain model caching and cleanup keys added to the list
    with patch.object(app_settings, "CACHE_DOMAIN_MODELS", True):
        cache = Redis(host=app_settings.CACHE_HOST, port=app_settings.CACHE_PORT)
        to_cleanup = []

        yield to_cleanup

        for key in to_cleanup:
            try:
                cache.delete(key)
            except Exception as exc:
                print("failed to delete cache key", key, str(exc))  # noqa: T001, T201


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_update_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that updating subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    cust_desc = make_customer_description(
        subscription_id=subscription.subscription_id,
        customer_id=subscription.customer_id,
        description="Original description",
    )
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be as inserted
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"][0]["description"] == "Original description"

    # Update customer description
    body = {
        "id": cust_desc.id,
        "subscription_id": subscription.subscription_id,
        "customer_id": subscription.customer_id,
        "description": "Updated description",
    }
    response2 = test_client.put("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is updated in the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response2.json()["customer_descriptions"][0]["description"] == "Updated description"


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_delete_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that deleting subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    cust_desc = make_customer_description(
        subscription_id=subscription.subscription_id,
        customer_id=subscription.customer_id,
        description="Original description",
    )
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be as inserted
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"][0]["description"] == "Original description"

    # Delete customer description
    response2 = test_client.delete(f"/api/subscription_customer_descriptions/{cust_desc.id}")
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is removed from the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response2.json()["customer_descriptions"] == []


@pytest.mark.skipif(
    not getenv("AIOCACHE_DISABLE", "0") == "0", reason="AIOCACHE must be enabled for this test to do anything"
)
def test_cache_create_customer_description(
    test_client, generic_subscription_1, make_customer_description, cache_fixture
):
    """Check that creating subscription customer description is reflected in the cache."""
    subscription = SubscriptionModel.from_subscription(generic_subscription_1)
    extended_model = build_extended_domain_model(subscription)

    # Add domainmodel to cache
    to_redis(extended_model)
    cache_fixture.extend([f"domain:{generic_subscription_1}", f"domain:etag:{generic_subscription_1}"])

    # Retrieve domain-model, customer description should be empty
    response1 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert response1.json()["customer_descriptions"] == []

    # Create customer description
    body = {
        "subscription_id": subscription.subscription_id,
        "customer_id": subscription.customer_id,
        "description": "New description",
    }
    response2 = test_client.post("/api/subscription_customer_descriptions/", json=body)
    assert HTTPStatus.NO_CONTENT == response2.status_code

    # Check that description is updated in the domain-model
    response2 = test_client.get(URL("api/subscriptions/domain-model") / generic_subscription_1)
    assert len(response2.json()["customer_descriptions"]) > 0
    assert response2.json()["customer_descriptions"][0]["description"] == "New description"



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Callable, Iterator, TypeVar

from more_itertools import partition
from pydantic import BaseModel
from sqlalchemy.sql import expression

from orchestrator.db.database import BaseModel as SqlBaseModel
from orchestrator.db.database import SearchQuery
from orchestrator.db.filters import CallableErrorHander


class SortOrder(Enum):
    ASC = "asc"
    DESC = "desc"


class Sort(BaseModel):
    field: str
    order: SortOrder

    class Config:
        use_enum_values = True


GenericType = TypeVar("GenericType")
QueryType = SearchQuery


def generic_sorts_validate(
    valid_sort_dict: dict[str, str]
) -> Callable[[list[Sort]], tuple[Iterator[Sort], Iterator[Sort]]]:
    """Create generic validate sort factory that creates a validate function based on the valid sort dict.

    Args:
        - valid_sort_dict: dict of column names by valid sort keys
            - key: sort key.
            - value: column name.

    Returns function that takes sort parameters and returns a list of invalid and valid Sort items.
    """

    def validate_sort_items(sort_by: list[Sort]) -> tuple[Iterator[Sort], Iterator[Sort]]:
        return partition(lambda item: item.field in valid_sort_dict, sort_by)

    return validate_sort_items


def generic_apply_sorts(
    valid_sort_dict: dict[str, str], model: SqlBaseModel
) -> Callable[[QueryType, Iterator[Sort]], QueryType]:
    def _apply_sorts(query: QueryType, sort_by: Iterator[Sort]) -> QueryType:
        for item in sort_by:
            field = item.field.lower()
            sort_key = valid_sort_dict[field]

            if item.order == SortOrder.DESC.value:  # type: ignore
                query = query.order_by(expression.desc(model.__dict__[sort_key]))
            else:
                query = query.order_by(expression.asc(model.__dict__[sort_key]))
        return query

    return _apply_sorts


def generic_sort(
    valid_sort_dict: dict[str, str],
    model: SqlBaseModel,
) -> Callable[[QueryType, list[Sort], CallableErrorHander], QueryType]:
    valid_sort_keys = list(valid_sort_dict.keys())
    _validate_sorts = generic_sorts_validate(valid_sort_dict)
    _apply_sorts = generic_apply_sorts(valid_sort_dict, model)

    def _sort(
        query: QueryType,
        sort_by: list[Sort],
        handle_sort_error: CallableErrorHander,
    ) -> QueryType:
        if sort_by:
            invalid_sort_items, valid_sort_items = _validate_sorts(sort_by)
            if invalid_list := [item.dict() for item in invalid_sort_items]:
                handle_sort_error(
                    "Invalid sort arguments",
                    invalid_filters=invalid_list,
                    valid_filter_keys=valid_sort_keys,
                )

            query = _apply_sorts(query, valid_sort_items)
        return query

    return _sort
